/*****************************************************
Chip type               : ATmega8
AVR Core Clock frequency: 16,000000 MHz
*****************************************************/
#include <mega8.h>
#include <math.h>
#include <delay.h>

#define A PORTC.2
#define B PORTC.3                 
#define C PORTC.4
#define D PORTC.5
#define E PORTD.0
#define F PORTD.1
#define G PORTD.2
#define DP PORTD.3

#define A1 PORTD.4
#define A2 PORTD.5
#define A3 PORTD.6

#define B1 PORTB.0
#define B2 PORTB.1
#define B3 PORTB.2


volatile unsigned char data [6];
volatile unsigned char counter = 0;

volatile unsigned char ADC [8];
volatile unsigned int volts = 0; 
volatile unsigned int amps = 0;
volatile unsigned int timer_ticks_u = 0;
volatile unsigned int timer_ticks_i = 0;
volatile unsigned int i_time = 256;

volatile unsigned int U;
volatile unsigned int I;
volatile unsigned int tmp_U;
volatile unsigned int tmp_I;

volatile signed int Ri = 0, Ru = 0;

//___________________________________________________________________________________
void ADC_Init()
{
  ADMUX = (1<<REFS1)|(1<<REFS0)|(1<<ADLAR)|(0<<MUX3)|(0<<MUX2)|(0<<MUX1)|(0<<MUX0);// Vref2,5V, left adst, cnl0
  ADCSR = (1<<ADEN)|(1<<ADSC)|(0<<ADFR)|(0<<ADIE)|(0<<ADPS2)|(0<<ADPS1)|(0<<ADPS0); //adc on 
}
//___________________________________________________________________________________
void port_init ()
{
DDRB = 0b00111111;
PORTB = 0x00;

DDRD = 0b01111111;
PORTD = 0x00;

DDRC = 0b00111100;
PORTC = 0x00;
}
//___________________________________________________________________________________
//void ACMP_init () 
//{ACSR = 0b01001010; }    // |= (1<<ACBG)|(1<<ACIE)|(1<<ACIS1)
//___________________________________________________________________________________
void TCCR0_init ()
{
TCCR0=0x03;  
TCNT0=0x00;
TIMSK |= 1<<TOIE0;
}
//___________________________________________________________________________________
void TCCR2_init ()
{
TCCR2=0b01101001; //clear on compare, 1:1, fast PWM
TCNT2=0x00;
OCR2=0x00;
}
//___________________________________________________________________________________
void  calculate_volts ()
{
   if (timer_ticks_u) {volts = ((unsigned long)Ru * 2600)/256;}
   else {volts = ((unsigned long) U * 2600)/256;}
      
}
//___________________________________________________________________________________ 
void  calculate_amps ()
{
   if (timer_ticks_i) {amps = ((unsigned long)Ri * 1000)/256;}
   else {amps = ((unsigned long)I * 1000)/256;}
}
//___________________________________________________________________________________
void ADC_cnl (char n)
{
 switch (n)
 {                            
 case 0: ADMUX = 0b11100000; break;
 case 1: ADMUX = 0b11100001; break;
 case 2: break;
 case 3: break;
 case 4: break;
 case 5: break;
 case 6: ADMUX = 0b11100110; break;
 case 7: ADMUX = 0b11100111; break;
 default : break; 
 }  
}
//___________________________________________________________________________________
void define_res ()
{
unsigned int tmp;

tmp = ADC [0];
if (((Ru - 25)>tmp)||((Ru+25)<tmp)) {timer_ticks_u = 3000;}
if (timer_ticks_u) {Ru = tmp;}

tmp = ADC [1];
if (((Ri - 50)>tmp)||((Ri+50)<tmp)) {timer_ticks_i = 3000;}
if (timer_ticks_i ) {Ri = tmp;}  
}
//___________________________________________________________________________________
void ADC_read ()
{
static unsigned char cnl = 0;
static unsigned char cnt = 0;

 ADC [cnl] = ADCH; 

 cnl ++;
 if (cnl == 2) {cnl = 6;} 
 if (cnl > 7) {cnl = 0;}
 if ((cnl == 0)&&(cnt < 100)) {cnl = 6;}
 if (cnt == 103) {cnl = 6; cnt = 0;}
 cnt++;
 ADC_cnl (cnl);
 define_res ();
 
 ADCSR |= (1<<ADSC);
 

}
//___________________________________________________________________________________
void stab_process ()
{ 
 if ((Ri > ADC [7])&&(ACSR.5))
 {
 if ((ADC [6] < Ru)&&(OCR2 < 255)) {OCR2++;}
 if ((ADC [6] > Ru)&&(OCR2 > 0)) {OCR2--;}
 }
else if (OCR2 > 0) {OCR2--;}
}
//___________________________________________________________________________________
void convert_data (unsigned int x, unsigned int y)
{
 unsigned int temp,res;
 temp=x;
 res=temp/1000;//Calculate 1000-s
 data[2]=res;
 temp=temp-res*1000;
 res=temp/100;//Calculate 100-s  
 data[1]=res;
 temp=temp-res*100;
 res=temp/10;//Calculate 10-s
 data[0]=res;

 temp=y;
 res=temp/1000;//Calculate 1000-s
 data[5]=res;
 temp=temp-res*1000;
 res=temp/100;//Calculate 100-s  
 data[4]=res;
 temp=temp-res*100;
 res=temp/10;//Calculate 10-s
 data[3]=res; 
 
}
//___________________________________________________________________________________
void clear_display ()
{
 A=1; B=1; C=1; D=1; E=1; F=1; G=1; DP=1; A1=0; A2=0; A3=0; B1=0; B2=0; B3=0;
}
//___________________________________________________________________________________
void write_code ()
{
switch (data[counter])
   {
      case 0:            (A=0),(B=0),(C=0),(D=0),(E=0),(F=0);           break;
      case 1:            (B=0),(C=0);                                                   break;
      case 2:            (A=0),(B=0),(D=0),(E=0),(G=0);                     break;
      case 3:            (A=0),(B=0),(C=0),(D=0),(G=0);                     break;
      case 4:            (B=0),(C=0),(F=0),(G=0);                               break;
      case 5:            (A=0),(C=0),(D=0),(F=0),(G=0);                     break;
      case 6:            (A=0),(C=0),(D=0),(E=0),(F=0),(G=0);           break;
      case 7:            (A=0),(B=0),(C=0);                                         break;
      case 8:            (A=0),(B=0),(C=0),(D=0),(E=0),(F=0),(G=0); break;
      case 9:            (A=0),(B=0),(C=0),(D=0),(F=0),(G=0);           break;
      default:          (G=0);                                                             break;
   }
} 
//___________________________________________________________________________________
void enable_digit  ()
{
 switch (counter)   
    {
     case 0: (A1 = 1); break;
     case 1: (A2 = 1), (DP=0); break;
     case 2: if (data[2]) {(A3 = 1);} else {(A3 = 0);}  break;  
     case 3: (B1 = 1); break;
     case 4: (B2 = 1), (DP=0); break;
     case 5: if (data[5]) {(B3 = 1);} else {(B3 = 0);}  break;
     default:               break;
    } 
}
//___________________________________________________________________________________
interrupt [TIM0_OVF] void timer0_ovf_isr(void)
{
 if (i_time) {tmp_U += ADC [6]; tmp_I += ADC [7];  i_time--;}
 else 
 {
 U = (tmp_U >> 8);
 tmp_U = 0;
   
 I =  (tmp_I >> 8);
 tmp_I = 0;
  
 calculate_volts ();
 calculate_amps ();
 convert_data (volts, amps); 
 i_time = 256;
 }
 
 clear_display ();
 enable_digit ();
 write_code ();
 if (counter < 7) {counter++;}
 else {counter = 0;}
 
 if (timer_ticks_u) {timer_ticks_u--;}
 if (timer_ticks_i) {timer_ticks_i--;} 
 

TCNT0 = 0;
}
//___________________________________________________________________________________
/*
interrupt [ANA_COMP] void ana_comp_isr(void)
{
 DDRB.3 = 0;
 OCR2 = 0;
 PORTB.5 = 1; 
 delay_ms (1000);
 PORTB.5 = 0;
 OCR2 = 0; 
 DDRB.3 = 1;
 }
*/
//___________________________________________________________________________________ 

void main(void)
{
port_init ();
ADC_Init();
//ACMP_init ();
TCCR0_init ();
TCCR2_init ();

#asm("sei")

while (1)
   {
    if (ADCSR.6 == 0) {ADC_read ();};
    stab_process ();   
   }
}
  
   
